"""
Parse string in to FileNamePart define in namer_types.
"""
import argparse
import logging
from pathlib import PurePath
import re
import sys
from typing import List
from namer.types import FileNameParts


DEFAULT_REGEX_TOKENS = '{_site}{_sep}{_date}{_sep}{_ts}{_name}{_dot}{_ext}'

def name_cleaner(name: str) -> str:
    """
    Given the name parts, following a date, but preceding the file extension, attempt to glean
    extra information and discard useless information for matching with the porndb.
    """
    # truncating cruft
    for size in ['2160p', '1080p', '720p', '4k',  '3840p']:
        name = re.sub(r"[\.\- ]"+size+r"[\.\- ]{0,1}.*", "", name)
    # remove trailing ".XXX."
    name = re.sub(r"[\.\- ]{0,1}XXX[\.\- ]{0,1}.*$", "", name)
    name = re.sub(r'\.', ' ', name)
    return name


def parser_config_to_regex(tokens: str) -> str:
    # pylint: disable=anomalous-backslash-in-string
    """
    ``{_site}{_sep}{_date}{_sep}{_ts}{_name}{_dot}{_ext}``

    ``Site - YYYY.MM.DD - TS - name.mkv``

    ```
    _sep    r'[\.\- ]+'
    _site   r'(?P<site>[a-zA-Z0-9\.\-\ ]+[a-zA-Z0-9])'
    _date   r'(?P<year>[0-9]{2}(?:[0-9]{2})?)[\.\- ]+(?P<month>[0-9]{2})[\.\- ]+(?P<day>[0-9]{2})'
    _ts     r'(?P<trans>[T|t][S|s])'+_sep+'){0,1}'
    _name   r'(?P<name>.*)'
    _dot    r'\.'
    _ext    r'(?P<ext>[a-zA-Z0-9]{3,4})$'
    ```
    """
    # pylint: enable=anomalous-backslash-in-string

    _sep=r'[\.\- ]+'
    _site=r'(?P<site>[a-zA-Z0-9\.\-\ ]+[a-zA-Z0-9])'
    _date=r'(?P<year>[0-9]{2}(?:[0-9]{2})?)[\.\- ]+(?P<month>[0-9]{2})[\.\- ]+(?P<day>[0-9]{2})'
    _ts=r'((?P<trans>[T|t][S|s])'+_sep+'){0,1}'
    _name=r'(?P<name>.*)'
    _dot=r'\.'
    _ext=r'(?P<ext>[a-zA-Z0-9]{3,4})$'
    regex = tokens.format_map({'_site':_site, '_date':_date, '_ts':_ts, '_name':_name, '_ext':_ext, '_sep':_sep, '_dot':_dot})
    return regex


def parse_file_name(filename: str, regex_config: str = DEFAULT_REGEX_TOKENS) -> FileNameParts:
    """
    Given an input name of the form site-yy.mm.dd-some.name.part.1.XXX.2160p.mp4,
    parses out the relevant information in to a structure form.
    """
    regex = parser_config_to_regex(regex_config)
    file_name_parts = FileNameParts()
    file_name_parts.extension = PurePath(filename).suffix[1:]
    match = re.search(regex,filename)
    if match:
        if match.groupdict().get('year') is not None:
            prefix = "20" if len(match.group('year'))==2 else ""
            file_name_parts.date = prefix+match.group('year')+"-"+match.group('month')+"-"+match.group('day')
        if match.groupdict().get('name') is not None:
            file_name_parts.name = name_cleaner(match.group('name'))
        if match.groupdict().get('site') is not None:
            file_name_parts.site = re.sub(r'[\.\-\ ]','',match.group('site'))
        if match.groupdict().get('trans') is not None:
            trans = match.group('trans')
            file_name_parts.trans = (not trans is None) and (trans.strip().upper() == 'TS')
        file_name_parts.extension = match.group('ext')
        file_name_parts.source_file_name = filename
    else:
        logging.warning("Could not parse file name: %s", filename)
    return file_name_parts


def main(arglist: List[str]):
    """
    Attempt to parse a name.
    """
    description = ("You are using the file name parser of the Namer project.  " +
        "Expects a single input, and will output the contents of FileNameParts, which is the internal input " +
        "to the namer_metadatapi.py script. "+
        "Output will be the representation of that FileNameParts.\n")
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("-f", "--file", help="String to parse for name parts", required=True)
    args = parser.parse_args(arglist)
    print(parse_file_name(args.file, DEFAULT_REGEX_TOKENS))

if __name__ == "__main__":
    main(arglist=sys.argv[1:])



"""
Updates mp4 files with metadata tags readable by Plex and Apple TV App.
"""
import logging
from pathlib import Path
from mutagen.mp4 import MP4, MP4Cover
from namer.types import LookedUpFileInfo, NamerConfig
from namer.ffmpeg import get_resolution, update_audio_stream_if_needed

logger = logging.getLogger('metadata')

def resolution_to_hdv_setting(resolution: int) -> int:
    """
    Using the resolution (height) of an video stream return the atom value for hdvideo
    """
    if resolution >= 2160:
        return 3
    if resolution >= 1080:
        return 2
    if resolution >= 720:
        return 1
    return 0

def update_mp4_file(mp4: Path, looked_up: LookedUpFileInfo, poster: Path, config: NamerConfig):
    # pylint: disable=too-many-statements
    """
    us-tv|TV-MA|600|
    us-tv|TV-14|500|
    us-tv|TV-PG|400|
    us-tv|TV-G|300|
    us-tv|TV-Y|200|
    us-tv|TV-Y7|100|
    us-tv||0|

    mpaa|UNRATED|600|
    mpaa|NC-17|500|
    mpaa|R|400|
    mpaa|PG-13|300|
    mpaa|PG|200|
    mpaa|G|100|
    mpaa|XXX|0|
    """

    logger.info("Updating audio and tags for: %s",mp4)
    success = update_audio_stream_if_needed(mp4, config.language)
    if not success:
        logger.info("Could not process audio or copy %s",mp4)
    logger.info("Updating atom tags on: %s",mp4)
    if mp4 is not None and mp4.exists():
        video = MP4(mp4)
        video["\xa9nam"] = [looked_up.name]
        video["\xa9day"] = [looked_up.date+"T09:00:00Z"]
        if config.enable_metadataapi_genres:
            video["\xa9gen"] = looked_up.tags
        else:
            video["keyw"] = looked_up.tags
            video["\xa9gen"] = [config.default_genre]
        video["tven"] = [looked_up.date]
        video["tvnn"] = [looked_up.site]
        video["stik"] = [9] #Movie
        video["hdvd"] = [resolution_to_hdv_setting(get_resolution(mp4))]
        video["ldes"] = [looked_up.description]
        video["----:com.apple.iTunes:iTunEXTC"] = 'mpaa|XXX|0|'.encode("UTF-8", errors="ignore")
        video["\xa9alb"] = [looked_up.site]
        itunes_movie = '<?xml version="1.0" encoding="UTF-8"?><plist version="1.0"><dict>'
        itunes_movie += f'<key>copy-warning</key><string>{looked_up.source_url}</string>'
        itunes_movie += f'<key>studio</key> <string>{looked_up.site}</string>'
        itunes_movie += f'<key>tpdbid</key> <string>{looked_up.look_up_site_id}</string>'
        itunes_movie += '<key>cast</key> <array>'
        for performer in looked_up.performers:
            if performer.name:
                itunes_movie += f'<dict> <key>name</key> <string>{performer.name}</string>'
                if performer.role:
                    itunes_movie += f'<key>role</key> <string>{performer.role}</string>'
                itunes_movie += '</dict>'
        itunes_movie += '</array>'
        itunes_movie += '<key>codirectors</key> <array></array>'
        itunes_movie += '<key>directors</key> <array></array>'
        itunes_movie += '<key>screenwriters</key><array></array>'
        itunes_movie += '</dict></plist>'
        video["----:com.apple.iTunes:iTunMOVI"] = itunes_movie.encode("UTF-8", errors="ignore")
        if poster is not None:
            with open(poster, "rb") as file:
                ext = poster.suffix.upper()
                imageformat = None
                if ext in ['.JPEG', '.JPG']:
                    imageformat=MP4Cover.FORMAT_JPEG
                elif ext == '.PNG':
                    imageformat=MP4Cover.FORMAT_PNG
                if imageformat:
                    video["covr"] = [
                        MP4Cover(file.read(), imageformat)
                    ]
        video.save()
        logger.info("Updated atom tags: %s", mp4)
    else:
        logger.warning("Can not update tags of a non-existant file: %s", mp4)



"""
This file can process individual movie files in place.
There name, or directory name, will be analyzed, matched against
the porndb, and used for renaming (in place), and updating an mp4
file's metadata (poster, artists, etc.)
"""
import argparse
import os
from pathlib import Path
import pathlib
import sys
import logging
from typing import List
from namer.moviexml import parse_movie_xml_file

from namer.types import LookedUpFileInfo, NamerConfig, ComparisonResult, ProcessingResults, default_config, from_config
from namer.filenameparser import parse_file_name
from namer.mutagen import update_mp4_file
from namer.metadataapi import get_poster, match

logger = logging.getLogger('namer')

DESCRIPTION="""
    Namer, the porndb local file renamer. It can be a command line tool to rename mp4/mkvs and to embed tags in mp4s,
    or a watchdog service to do the above watching a directory for new files and moving matched files to a target location.
    File names are assumed to be of the form SITE.[YY]YY.MM.DD.String.of.performers.and.or.scene.name.<IGNORED_INFO>.[mp4|mkv].   In the name, read the
    periods, ".", as any number of spaces " ", dashes "-", or periods ".".

    Provided you have an access token to the porndb (free sign up) https://www.metadataapi.net/, this program will
    attempt to match your file's name to search results from the porndb.   Please note that the site must at least be
    a substring of the actual site name on the porndb, and the date must be within one day or the release date on the
    porndb for a match to be considered.  If the log file flag is enabled then a <original file name minus ext>_namer.log
    file will be written with all the potential matches sorted, descending by how closely the scene name/performer names
    match the file's name segment after the 'SITE.[YY]YY.MM.DD'.
  """

def write_log_file(movie_file: Path, match_attempts: List[ComparisonResult]) -> str:
    """
    Given porndb scene results sorted by how closely they match a file,  write the contents
    of the result matches to a log file.
    """
    logname = movie_file.with_name(movie_file.stem+"_namer.log")
    logger.info("Writing log to %s",logname)
    with open(logname, "wt", encoding='utf-8') as log_file:
        for attempt in match_attempts:
            log_file.write("\n")
            log_file.write(f"File                : {attempt.name_parts.source_file_name}\n")
            log_file.write(f"Scene Name          : {attempt.looked_up.name}\n")
            log_file.write(f"Match               : {attempt.is_match()}\n")
            log_file.write(f"Query URL           : {attempt.looked_up.original_query}\n")
            if attempt.name_parts.site is None:
                attempt.name_parts.site = 'None'
            if attempt.name_parts.date is None:
                attempt.name_parts.date = 'None'
            if attempt.name_parts.date is None:
                attempt.name_parts.name = 'None'
            log_file.write(f"{str(attempt.sitematch):5} Found Sitename: {attempt.looked_up.site:50.50} Parsed Sitename:"+
                f" {attempt.name_parts.site:50.50}\n")
            log_file.write(f"{str(attempt.datematch):5} Found Date    : {attempt.looked_up.date:50.50} Parsed Date    :"+
                f" {attempt.name_parts.date:50.50}\n")
            log_file.write(f"{attempt.name_match:5.1f} Found Name    : {attempt.name:50.50} Parsed Name    :"+
                f" {attempt.name_parts.name:50.50}\n")
    return logname

def set_permissions(file: Path, config: NamerConfig):
    """
    Given a file or dir, set permissions from NamerConfig.set_file_permissions,
    NamerConfig.set_dir_permissions, and uid/gid if set for the current process recursively.
    """
    if hasattr(os, "chmod") and file is not None and file.exists():
        fileperm: int = None if config.set_file_permissions is None else int(str(config.set_file_permissions), 8)
        dirperm: int = None if config.set_dir_permissions is None else int(str(config.set_dir_permissions), 8)
        if dirperm is not None and file.is_dir():
            Path(file).chmod(dirperm)
            if config.set_uid is not None and config.set_gid is not None:
                os.chown(file, uid=config.set_uid, gid=config.set_gid)
        if file.is_dir():
            for root, _dirs, files in os.walk(file):
                if dirperm is not None:
                    Path(root).chmod(dirperm)
                if config.set_uid is not None and config.set_gid is not None:
                    os.chown(root, uid=config.set_uid, gid=config.set_gid)
                for cur in files:
                    set_permissions(Path(os.path.join(root, cur)), config)
        elif file.is_file() and fileperm is not None:
            if fileperm is not None:
                file.chmod(fileperm)
                if config.set_uid is not None and config.set_gid is not None:
                    os.chown(file, uid=config.set_uid, gid=config.set_gid)


def dir_with_subdirs_to_process(dir_to_scan: Path, config: NamerConfig, infos: bool = False):
    """
    Used to find subdirs of a directory to be individually processed.
    The directories will be scanned for media and named/tagged in place
    based on config settings.
    """
    if dir_to_scan is not None and dir_to_scan.is_dir() and dir_to_scan.exists():
        logger.info("Scanning dir %s for subdirs/files to process",dir_to_scan)
        files = list(dir_to_scan.iterdir())
        files.sort()
        for file in files:
            fullpath_file = dir_to_scan / file
            if fullpath_file.is_dir() or fullpath_file.suffix.upper() in [".MP4",".MKV"]:
                process(fullpath_file, config, infos)

def tag_in_place(video: Path, config: NamerConfig, new_metadata: LookedUpFileInfo):
    """
    Uses ComparisonResults to update an mp4 file's metadata based on a match in
    ComparisonResults.   Expects the first item of list to be the match if there is one.
    Will download a poster as well depending on NamerConfig config setting.
    """
    if new_metadata is not None:
        poster = None
        if config.enabled_tagging is True and video.suffix.lower() == ".mp4":
            if config.enabled_poster is True:
                logger.info("Downloading poster: %s",new_metadata.poster_url)
                poster = get_poster(new_metadata.poster_url, config.porndb_token, video)
                set_permissions(poster, config)
            logger.info("Updating file metadata (atoms): %s",video)
            update_mp4_file(video, new_metadata, poster, config)
        logger.info("Done tagging file: %s",video)
        if poster is not None and new_metadata.poster_url.startswith("http"):
            poster.unlink()


def find_largest_file_in_glob(rootdir: Path, globstr: str) -> Path:
    """
    returns largest matching file
    """
    list_of_files = list(rootdir.rglob(globstr))
    logger.info("found files %s", list_of_files)
    file = None
    if len(list_of_files) > 0:
        file = max( list_of_files, key =  lambda x: x.stat().st_size)
    return file


def determine_target_file(file_to_process: Path, config: NamerConfig) -> ProcessingResults:
    """
    Base on the file to process - which may be a file or a dir, and configuration, determine
    the file if needed (largest mp4, or mkv in a directory), or the directory (parent dir of file),
    as well as determine which should be used for attempted naming.

    If config.prefer_dir_name_if_available is set to True and a directory was passed as file_to_process
    then the dirctory's name will be returned as name, else the found file's name is used.
    """
    results = ProcessingResults()

    containing_dir = None
    if file_to_process.is_dir():
        logger.info("Target dir: %s",file_to_process)
        containing_dir = file_to_process
        file = find_largest_file_in_glob(file_to_process, "**/*.mp4")
        if file is None:
            file = find_largest_file_in_glob(file_to_process, "**/*.mkv")
    else:
        file = file_to_process

    if config.prefer_dir_name_if_available and containing_dir is not None:
        name = containing_dir.stem+file.suffix
    else:
        name = file.name

    logger.info("file: %s",file)
    logger.info("dir : %s",containing_dir)

    results.dirfile = containing_dir
    results.video_file = file
    results.parsed_file = parse_file_name(name, config.name_parser)
    if containing_dir is True:
        results.final_name_relative = file.relative_to(containing_dir)
    else:
        results.final_name_relative = file.parent
    return results

def get_local_metadata_if_requested(video_file: Path) -> LookedUpFileInfo:
    """
    If there is an .nfo file next to the video_file, attempt to read it as
    a Emby/Jellyfin style movie xml file.
    """
    nfo_file = video_file.parent / (video_file.stem + ".nfo")
    if nfo_file.is_file() and nfo_file.exists():
        return parse_movie_xml_file(nfo_file)
    return None

def move_to_final_location(to_move: Path, target_dir: Path, template: str, new_metadata: LookedUpFileInfo) -> Path:
    """
    Moves a file or directory to it's final location after verifying there is no collision.
    Should a collision occur, the file is appropriately renamed to avoid collision.
    """
    infix = 0
    while True:
        relative_path = new_metadata.new_file_name(template, f"({infix})")
        newname = target_dir / relative_path
        newname = newname.resolve()
        infix += 1
        if not newname.exists() or to_move.samefile(newname):
            break
    os.makedirs(newname.parent,exist_ok=True)
    to_move.rename(newname)
    return newname


def process(file_to_process: Path, config: NamerConfig, infos: bool = False) -> ProcessingResults:
    """
    Bread and butter method.
    Given a file, determines if it's a dir, if so, the dir name may be used
    for comparison with the porndb.   The larges mp4/mkv file in the directory
    or any subdirectories will be assumed to be the movie file.

    Does not properly handle multi-part movies.

    If the input file is not a dir it's name will be used and it is assumed to
    be the movie file we wish to tag.

    The movie is either renamed in place if a file, or renamed and move to the root
    of the dir if a dir was passed in.

    The file is then update based on the metadata from the porndb if an mp4.
    """
    logger.info("Analyzing: %s",file_to_process)
    output: ProcessingResults = determine_target_file(file_to_process, config)
    if output.video_file is not None:
        logger.info("Processing: %s", output.video_file)
        # Match to nfo files, if enabled and found.
        if infos is True:
            output.new_metadata = get_local_metadata_if_requested(output.video_file)
            if output.new_metadata is not None:
                output.new_metadata.original_parsed_filename = output.parsed_file
        if (output.new_metadata is None and
            output.parsed_file is not None and
            output.parsed_file.name is not None):
            output.search_results = match(output.parsed_file, config.porndb_token)
            if len(output.search_results) > 0 and output.search_results[0].is_match() is True:
                output.new_metadata = output.search_results[0].looked_up
        else:
            logger.error(
            "Could not parse file: %s, it is not in the right format it must start with a site, a date and end with an extension",
            output.video_file)
        target_dir = output.dirfile if output.dirfile is not None else output.video_file.parent
        set_permissions(target_dir, config)
        if output.new_metadata is not None:
            output.video_file = move_to_final_location(
                to_move=output.video_file,
                target_dir=output.video_file.parent,
                template=config.inplace_name,
                new_metadata=output.new_metadata)
            set_permissions(output.video_file, config)
            tag_in_place(output.video_file, config, output.new_metadata)
            logger.info("Done processing file: %s, moved to %s", file_to_process,output.video_file)
        else:
            #not matched.
            output.final_name_relative=os.path.relpath(output.video_file, target_dir)

        # Write log file if needed.
        if config.write_namer_log is True and output.search_results is not None:
            logfile = write_log_file(output.video_file, output.search_results)
            set_permissions(logfile, config)
            output.namer_log_file=logfile
    return output

def check_arguments(file_to_process: Path, dir_to_process: Path, config_overide: Path):
    """
    check arguments.
    """
    error = False
    if file_to_process is not None:
        logger.info("File to process: %s", file_to_process)
        if not file_to_process.is_file() or not file_to_process.exists():
            logger.error("Error not a file! %s", file_to_process)
            error = True

    if dir_to_process is not None:
        logger.info("Directory to process: %s",dir_to_process)
        if not dir_to_process.is_dir() or not dir_to_process.exists():
            logger.info("Error not a directory! %s", dir_to_process)
            error = True

    if config_overide is not None:
        logger.info("Config override specified: %s",config_overide)
        if not config_overide.is_file() or not config_overide.exists():
            logger.info("Config override specified, but file does not exit: %s",config_overide)
            error = True
    return error

def main(arglist: List[str]):
    """
    Used to tag and rename files from the command line.
    See usage function above.
    """
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("-c", "--configfile", help="config file, defaults first to env var NAMER_CONFIG,"
        +" then local path namer.cfg, and finally ~/.namer.cfg.", type=pathlib.Path)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-f", "--file", help="a single file to process, and rename.", type=pathlib.Path)
    group.add_argument("-d", "--dir", help="a directory to process.", type=pathlib.Path)
    parser.add_argument("-m", "--many", help="if set, a directory have all it's sub directories processed."+
        " Files move only within sub dirs, or are renamed in place, if in the root dir to scan", action="store_true")
    parser.add_argument("-i", "--infos", help="if set, .nfo files will attempt to be accessed next to movie files" +
        ", if info files are found and parsed successfully, that metadata will be used rather than " +
        "porndb matching.  If using jellyfin .nfo files, please bump your release date by one day "+
        "until they fix this issue: https://github.com/jellyfin/jellyfin/issues/7271.", action="store_true")
    parser.add_argument("-v", "--verbose", help="verbose, print logs", action="store_true")
    args = parser.parse_args(arglist)
    logger_level = logging.ERROR
    if args.verbose:
        logger_level=logging.DEBUG
    logging.basicConfig(level=logger_level)
    check_arguments(args.file, args.dir, args.configfile)

    config = default_config()
    if args.configfile is not None and args.configfile.is_file():
        logger.info("Config override specified %s",args.configfile)
        config = from_config(args.configfile)
    config.verify_naming_config()
    target = args.file
    if args.dir is not None:
        target = args.dir
    if args.many is True:
        dir_with_subdirs_to_process(args.dir, config, args.infos)
    else:
        process(target, config, args.infos)


if __name__ == "__main__":
    main(sys.argv[1:])



"""
A file watching service to rename movie files and move them
to revelant locations after match the file against the porndb.
"""

import shutil
import tempfile
import time
import os
import sys
import traceback
from pathlib import Path, PurePath
import logging
from watchdog.observers.polling import PollingObserver
from watchdog.events import PatternMatchingEventHandler, FileSystemEvent, EVENT_TYPE_DELETED, EVENT_TYPE_MOVED
import schedule
from namer.namer import move_to_final_location, process
from namer.types import NamerConfig, default_config

logger = logging.getLogger('watchdog')


def done_copying(file: Path) -> bool:
    """
    Determines if a file is being copied by checking it's size in 2 second
    increments and seeing if the size has stayed the same.
    """
    size_past = 0
    while True:
        if file is None or not file.exists():
            return False
        size_now = file.stat().st_size
        if size_now == size_past:
            return True
        size_past = size_now
        time.sleep(.2)


def handle(target_file: Path, namer_config: NamerConfig):
    """
    Responsible for processing and moving new movie files.
    """
    relative_path = target_file.relative_to(namer_config.watch_dir)

    # is in a dir:
    detected = PurePath(relative_path).parts[0]
    dir_path = namer_config.watch_dir / detected

    to_process = None
    workingdir = None
    workingfile = None
    if dir_path.is_dir():
        workingdir = Path(namer_config.work_dir) / detected
        logger.info("Moving %s to %s for processing", dir_path, workingdir)
        shutil.move(dir_path,workingdir)
        to_process = workingdir
    else:
        workingfile = Path(namer_config.work_dir) / relative_path
        target_file.rename(workingfile)
        logger.info("Moving %s to %s for processing", target_file, workingfile)
        to_process = workingfile
    result = process(to_process, namer_config)

    if result.new_metadata is None:
        if workingdir is not None:
            workingdir.rename(namer_config.failed_dir/ detected)
            logger.info("Moving failed processing %s to %s to retry later", workingdir, namer_config.failed_dir/ detected)
        else:
            newvideo = namer_config.failed_dir / relative_path
            workingfile.rename(newvideo)
            logger.info("Moving failed processing %s to %s to retry later", workingfile, newvideo)
            if result.namer_log_file is not None:
                result.namer_log_file.rename( newvideo.parent / result.namer_log_file.name)
    else:
        # See if we should, and can, move the whole dir.
        moved = False
        if (
            len(PurePath(result.final_name_relative).parts) > 1
            and workingdir is not None
            and namer_config.del_other_files is False
        ):
            target = namer_config.dest_dir / PurePath(result.final_name_relative).parts[0]
            if not target.exists():
                shutil.move(workingdir, target)
                moved = True
                logger.info("Moving success processed dir %s to %s", workingdir, target)
        # else just moved the tagged video file and logs.
        if not moved:
            newfile = move_to_final_location(
                result.video_file,
                namer_config.dest_dir,
                namer_config.new_relative_path_name,
                result.new_metadata)
            if result.namer_log_file is not None:
                shutil.move(result.namer_log_file, newfile.parent / (newfile.stem+"_namer.log"))
            logger.info("Moving success processed file %s to %s", workingfile, newfile)
            if workingdir is not None:
                shutil.rmtree(workingdir, ignore_errors=True)

def retry_failed(namer_config: NamerConfig):
    """
    Moves the contents from the failed dir to the watch dir to attempt reprocessing.
    """
    logger.info("Retry failed items:")
    for file in list(namer_config.failed_dir.iterdir()):
        shutil.move( namer_config.failed_dir / file.name, namer_config.watch_dir / file.name )

def is_fs_case_sensitive():
    """
    Create a temporary file to determine if a filesystem is case sensitive, or not.
    """
    with tempfile.NamedTemporaryFile(prefix='TmP') as tmp_file:
        return(not os.path.exists(tmp_file.name.lower()))

class MovieEventHandler(PatternMatchingEventHandler):
    """
    When a new movie file is detected, this class handles the event,
    and passes off the processing to the handler() function above.
    """
    namer_config: NamerConfig

    def __init__(self, namer_config: NamerConfig):
        super().__init__(patterns=["**/*.mp4", "**/*.mkv", "**/*.MP4", "**/*.MKV"],
                         case_sensitive=is_fs_case_sensitive(), ignore_directories=True, ignore_patterns=None)
        self.namer_config = namer_config

    def on_any_event(self, event: FileSystemEvent):
        file_path = None
        if event.event_type == EVENT_TYPE_MOVED:
            file_path = event.dest_path
        elif event.event_type != EVENT_TYPE_DELETED:
            file_path = event.src_path
        if file_path is not None:
            path = Path(file_path)
            logger.info("watchdog process called for %s", path)
            if path.exists() and done_copying(path) and (path.stat().st_size / (1024*1024) > self.namer_config.min_file_size):
                # Extra wait time in case other files are copies in as well.
                if self.namer_config.del_other_files is True:
                    time.sleep(self.namer_config.extra_sleep_time)
                try:
                    handle(path, self.namer_config)
                except Exception as ex:  # pylint: disable=broad-except
                    try:
                        exc_info = sys.exc_info()
                        try:
                            logger.error("Error handling %s: \n %s", path, ex)
                        except Exception: # pylint: disable=broad-except
                            pass
                    finally:
                        # Display the *original* exception
                        traceback.print_exception(*exc_info)
                        del exc_info


class MovieWatcher:
    """
    Watches a configured dir for new files and attempts
    to process them and move to a destination dir.

    If a failure occures moves the new files to a failed dir.

    See NamerConfig
    """
    def __init__(self, namer_config: NamerConfig):
        self.__namer_config = namer_config
        self.__src_path = namer_config.watch_dir
        self.__event_handler = MovieEventHandler(namer_config)
        self.__event_observer = PollingObserver()

    def run(self):
        """
        Checks for new files in 3 second intervals,
        needed if running in docker as events aren't properly passed in.
        """
        self.start()
        try:
            while True:
                schedule.run_pending()
                time.sleep(3)
        except KeyboardInterrupt:
            self.stop()

    def start(self):
        """
        starts a background thread to check for files.
        """
        config = self.__namer_config
        logger.info("Start porndb scene watcher.... watching: %s",config.watch_dir)
        if os.environ.get('BUILD_DATE'):
            build_date = os.environ.get('BUILD_DATE')
            print(f"Built on: {build_date}")
        if os.environ.get('GIT_HASH'):
            git_hash = os.environ.get('GIT_HASH')
            print(f"Git Hash: {git_hash}")
        self.__schedule()
        self.__event_observer.start()
        # touch all existing movie files.
        for root, _dirnames, filenames in os.walk(self.__namer_config.watch_dir):
            for file in filenames:
                ext = os.path.splitext(file)[1].lower()
                if ext in {'.mkv', '.mp4'}:
                    movie = Path(os.path.join(root, file))
                    movie.touch()

    def stop(self):
        """
        stops a background thread to check for files.
        Waits for the event processor to complete.
        """
        logger.info("exiting")
        self.__event_observer.stop()
        self.__event_observer.join()
        logger.info("exited")

    def __schedule(self):
        self.__event_observer.schedule(
            self.__event_handler,
            self.__src_path,
            recursive=True
        )


def create_watcher(namer_watchdog_config: NamerConfig) -> MovieWatcher:
    """
    Configure and start a watchdog looking for new Movies.
    """
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')
    logger.info(str(namer_watchdog_config))
    if not namer_watchdog_config.verify_watchdog_config():
        sys.exit(-1)
    if namer_watchdog_config.retry_time is not None:
        schedule.every().day.at(namer_watchdog_config.retry_time).do(lambda: retry_failed(namer_watchdog_config))
    movie_watcher = MovieWatcher(namer_watchdog_config)
    return movie_watcher


if __name__ == "__main__":
    create_watcher(default_config()).run()



"""
Handle matching data in a FileNameParts (likely generated by a namer_file_parser.py) to
look up metadata (actors, studio, creation data, posters, etc) from the porndb.
"""

import argparse
import itertools
import os
from pathlib import Path
import json
from random import choices
import string
from datetime import timedelta, date
import pathlib
import re
import sys
from typing import List
import logging
from types import SimpleNamespace
import urllib
import urllib.request
import rapidfuzz
import requests
from namer.types import LookedUpFileInfo, Performer, FileNameParts, ComparisonResult, default_config
from namer.filenameparser import parse_file_name

logger = logging.getLogger('metadata')

def __evaluate_match(name_parts: FileNameParts, looked_up: LookedUpFileInfo) -> ComparisonResult:
    release_date = name_parts.date is None or name_parts.date == looked_up.date
    site = (name_parts.site is None or
        re.sub(r' ', '', name_parts.site.capitalize()) in re.sub( r' ', '', looked_up.site.capitalize()))
    all_performers = list(map(lambda p: p.name, looked_up.performers))
    all_performers.insert(0, looked_up.name)
    powerset = (combo for r in range(1,len(all_performers) + 1) for combo in itertools.combinations(all_performers, r))
    ratios = rapidfuzz.process.extractOne(name_parts.name, map(" ".join, powerset))
    return ComparisonResult(
        name=ratios[0],
        name_match=ratios[1],
        datematch=release_date,
        sitematch=site,
        name_parts=name_parts,
        looked_up=looked_up)

def __update_results(results: List[ComparisonResult], name_parts: FileNameParts, authtoken: str,
        skipdate: bool=False, skipname: bool=False):
    if len(results) == 0 or not results[0].is_match():
        for match_attempt in __get_metadataapi_net_fileinfo(name_parts, authtoken, skipdate, skipname):
            result = __evaluate_match(name_parts, match_attempt)
            results.append(result)
        results = sorted(results, key=__match_percent, reverse=True)

def __metadata_api_lookup(name_parts: FileNameParts, authtoken: str) -> List[ComparisonResult]:
    results = []
    __update_results(results, name_parts, authtoken)
    __update_results(results, name_parts, authtoken, skipdate=True)
    __update_results(results, name_parts, authtoken, skipdate=True, skipname=True)
    __update_results(results, name_parts, authtoken, skipname=True)

    if len(results) == 0 or not results[-1].is_match():
        name_parts.date =  (date.fromisoformat(name_parts.date)+timedelta(days=-1)).isoformat()
        logger.info("Not found, trying 1 day before: %s",name_parts)
        __update_results(results, name_parts, authtoken)
        __update_results(results, name_parts, authtoken, skipdate=False, skipname=True)

    if len(results) == 0 or not results[-1].is_match():
        name_parts.date = (date.fromisoformat(name_parts.date)+timedelta(days=2)).isoformat()
        logger.info("Not found, trying 1 day after: %s",name_parts)
        __update_results(results, name_parts, authtoken)
        __update_results(results, name_parts, authtoken, skipdate=False, skipname=True)
    return results


def __match_percent(result: ComparisonResult) -> float:
    addvalue=0.00
    if result.is_match() is True:
        addvalue=1000.00
    value = result.name_match + addvalue
    logger.info("Name match was %.2f for %s", value, result.name)
    return value


def __get_response_json_object(url: str, authtoken: str) -> str:
    """
    returns json object with info
    """
    headers = {"Authorization": f"Bearer {authtoken}",
               "Content-Type": "application/json",
               "Accept": "application/json",
               'User-Agent': 'namer-1'}
    try:
        with requests.get(url, headers=headers) as response:
            response.raise_for_status()
            return response.text
    except requests.exceptions.RequestException as ex:
        logger.warning(ex)
        return None

def get_poster(url: str, authtoken: str, video_file: Path) -> Path:
    """
    returns json object with info
    """
    if url.startswith("http"):
        headers = {"Authorization": f"Bearer {authtoken}",
                'User-Agent': 'namer-1'}
        random = ''.join(choices(population=string.ascii_uppercase + string.digits, k=10))
        file =  video_file.parent / ( video_file.stem + "_" + random+"_poster" + pathlib.Path(url).suffix )
        try:
            with requests.get(url, headers=headers) as response:
                #Not sure how to avoid this 406, tried all kinds of Accept/User-Agent...
                #response.raise_for_status()
                with open(file, "wb") as binary_file:
                    # Write bytes to file
                    binary_file.write(response.content)
                    return file
        except requests.exceptions.RequestException as ex:
            logger.warning(ex)
            return None
    else:
        poster = (video_file.parent / url).resolve()
        return poster if poster.exists() and poster.is_file() else None


def __jsondata_to_fileinfo(data, url, json_response, name_parts) -> LookedUpFileInfo:
    file_info = LookedUpFileInfo()
    file_info.uuid = data._id # pylint: disable=protected-access
    file_info.name = data.title
    file_info.description = data.description
    file_info.date = data.date
    file_info.source_url = data.url
    file_info.poster_url = data.poster
    file_info.site = data.site.name
    file_info.look_up_site_id = data._id # pylint: disable=protected-access
    for json_performer in data.performers:
        performer = Performer()
        if hasattr(json_performer, "parent") and hasattr(json_performer.parent, "extras"):
            performer.role = json_performer.parent.extras.gender
        performer.name = json_performer.name
        file_info.performers.append(performer)
    file_info.original_query=url
    file_info.origninal_response=json_response
    file_info.original_parsed_filename=name_parts
    tags = []
    if hasattr(data, "tags"):
        for tag in data.tags:
            tags.append(tag.name)
        file_info.tags = tags
    return file_info


def __metadataapi_response_to_data(json_object, url, json_response, name_parts) -> List[LookedUpFileInfo]:
    file_infos = []
    if hasattr(json_object, "data"):
        if isinstance(json_object.data, list):
            for data in json_object.data:
                found_file_info = __jsondata_to_fileinfo(data, url, json_response, name_parts)
                file_infos.append(found_file_info)
        else:
            file_infos.append(__jsondata_to_fileinfo(json_object.data, url, json_response, name_parts))
    return file_infos


def __build_url(site:str=None, release_date:str=None, name:str=None, uuid:str=None) -> str:
    query = ""
    if uuid is not None:
        query = "/"+str(uuid)
    else:
        query="?parse="
        if site is not None:
            query += urllib.parse.quote(re.sub(r' ', '.', site))+"."
        if release_date is not None:
            query += release_date+"."
        if name is not None:
            query += urllib.parse.quote(re.sub(r' ', '.', name))
        query+="&limit=25"
    return f"https://api.metadataapi.net/scenes{query}"


def __get_metadataapi_net_fileinfo(name_parts: FileNameParts, authtoken: str, skipdate: bool,
        skipname: bool) -> List[LookedUpFileInfo]:
    release_date = name_parts.date if not skipdate else None
    name = name_parts.name if not skipname else None
    url = __build_url(name_parts.site, release_date, name)
    logger.info("Querying: %s",url)
    json_response = __get_response_json_object(url, authtoken)
    file_infos = []
    if json_response is not None and json_response.strip() != '':
        logger.debug("json_repsonse: \n%s", json_response)
        json_obj = json.loads(json_response, object_hook=lambda d: SimpleNamespace(**d))
        formatted = json.dumps(json.loads(json_response), indent=4, sort_keys=True)
        file_infos = __metadataapi_response_to_data(json_obj, url, formatted, name_parts)
    return file_infos


def __get_complete_metadatapi_net_fileinfo(name_parts: FileNameParts, uuid: str, authtoken: str) -> LookedUpFileInfo:
    url = __build_url(uuid=uuid)
    logger.info("Querying: %s",url)
    json_response = __get_response_json_object(url, authtoken)
    file_infos = []
    if json_response is not None and json_response.strip() != '':
        logger.debug("json_repsonse: \n%s",json_response)
        json_obj = json.loads(json_response, object_hook=lambda d: SimpleNamespace(**d))
        formatted = json.dumps(json.loads(json_response), indent=4, sort_keys=True)
        file_infos = __metadataapi_response_to_data(json_obj, url, formatted, name_parts)
    if len(file_infos) > 0:
        return file_infos[0]
    return None

def match(file_name_parts: FileNameParts, porndb_token: str) -> List[ComparisonResult]:
    """
    Give parsed file name parts, and a porndb token, returns a sorted list of possible matches.
    Matches will appear first.
    """
    comparison_results = __metadata_api_lookup(file_name_parts, porndb_token)
    comparison_results = sorted(comparison_results, key=__match_percent, reverse=True)
    # Works around the porndb not returning all info on search queries by looking up the full data
    # with the uuid of the best match.
    if len(comparison_results) > 0 and comparison_results[0].is_match() is True:
        file_infos =  __get_complete_metadatapi_net_fileinfo(file_name_parts, comparison_results[0].looked_up.uuid ,porndb_token)
        if file_infos is not None:
            comparison_results[0].looked_up = file_infos
    return comparison_results


def main(argslist: List[str]):
    """
    Looks up metadata from metadataapi.net base on file name.
    """
    description = """
    Command line interface to look up a suggested name for an adult movie file based on an input string
    that is parsable by namer_file_parser.py
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("-c", "--configfile", help="override location for a configuraion file.", type=pathlib.Path)
    parser.add_argument("-f", "--file", help="File we want to provide a match name for.", required=True, type=pathlib.Path)
    parser.add_argument("-t", "--token", help="Access token.", required=True)
    parser.add_argument("-j", "--jsonfile", help="write returned json to this file.", type=pathlib.Path)
    parser.add_argument("-v", "--verbose", help="verbose, print logs", action="store_true")
    args = parser.parse_args(args=argslist)
    logger_level = logging.ERROR
    if args.verbose:
        logger_level=logging.DEBUG
    logging.basicConfig(level=logger_level)
    file_name = parse_file_name(os.path.basename(args.file), default_config().name_parser)
    match_results = match(file_name, args.token)
    if len(match_results) > 0 and match_results[0].is_match() is True:
        print(match_results[0].looked_up.new_file_name(default_config().inplace_name))
        if args.jsonfile is not None:
            Path(args.jsonfile).write_text(match_results[0].looked_up.origninal_response, encoding="UTF-8")

if __name__ == "__main__":
    main(sys.argv[1:])



"""
Namer, the porn db file renamer. It can be a command line tool to rename mp4/mkvs and to embed tags in mp4s,
or a watchdog service to do the above watching a directory for new files.  File names are assumed to be of
the form SITE.[YY]YY.MM.DD.String.of.performers.and.or.scene.name.<IGNORED_INFO>.[mp4|mkv].   In the name, read the
periods, ".", as any number of spaces " ", dashes "-", or periods ".".

Provided you have an access token to the porndb (free sign up) https://www.metadataapi.net/, this program will
attempt to match your file's name to search results from the porndb.   Please note that the site must at least be
a substring of the actual site name on the porndb, and the date must be within one day or the release date on the
porndb for a match to be considered.  If the log file flag is enabled then a <original file name minus ext>_namer.log
file will be written with all the potential matches sorted, descending by how closely the scene name/performer names
match the file.
"""
import pathlib
import sys
from typing import List
from namer.types import default_config
import namer.watchdog
import namer.namer

DESCRIPTION=namer.namer.DESCRIPTION+"""

    The first argument should be 'watchdog', 'rename', or 'help' to see this message, for more help on rename, call
    namer 'namer rename -h'

    watchdog and help take no arguments (please see the config file example https://github.com/4c0d3r/namer/blob/main/namer.cfg)
    """

def create_default_config_if_missing():
    """
    Find or create config.
    """
    conffile = pathlib.Path(".namer.conf")
    print("Creating default config file here: %s", conffile)
    print("please edit the token or any other settings whose defaults you want changed.")


def main(arglist: List[str]):
    """
    Call main method in namer.namer or namer.watchdog.
    """
    arg1 = None if len(arglist) == 0 else arglist[0]
    if arg1 == 'watchdog':
        namer.watchdog.create_watcher(default_config()).run()
    elif arg1 == 'rename':
        namer.namer.main(arglist[1:])
    elif arg1 in ['-h','help', None]:
        print(DESCRIPTION)

if __name__ == "__main__":
    main(sys.argv[1:])



"""
Reads movie.xml (your.movie.name.nfo) of Emby/Jellyfin format in to a LookedUpFileInfo,
allowing the metadata to be written in to video files (currently only mp4s),
or used in renaming the video file.
"""
from pathlib import Path
from lxml import objectify
from namer.types import LookedUpFileInfo, Performer

def parse_movie_xml_file(xmlfile: Path) -> LookedUpFileInfo:
    """
    Parse an Emby/Jellyfin xml file and creates a LookedUpFileInfo from the data.
    """
    content = xmlfile.read_text(encoding="utf8")

    movie = objectify.fromstring(bytes(content, encoding="utf8"))
    info = LookedUpFileInfo()
    info.name = str(movie.title)
    info.site = str(movie.studio[0])
    info.date = str(movie.releasedate)
    info.description = str(movie.plot)
    info.poster_url = str(movie.art.poster)
    info.performers = []
    for actor in movie.actor:
        performer = Performer()
        performer.name = str(actor.name)
        performer.role = str(actor.role)
        info.performers.append(performer)
    if hasattr(movie, "phoenixadulturlid"):
        info.look_up_site_id = str(movie.phoenixadulturlid)
    if hasattr(movie, "theporndbid"):
        info.uuid = str(movie.theporndbid)
    info.tags = []
    for genre in movie.genre:
        info.tags.append(str(genre))
    info.original_parsed_filename = None
    info.original_query = None
    info.origninal_response = None
    return info



"""
Types shared by all the files in this project, used as interfaces for moving data.
"""

from configparser import ConfigParser
from dataclasses import dataclass
from typing import List
import os
import configparser
import re
import sys
import string
from pathlib import Path, PurePath
import logging
import random
from pathvalidate import Platform, sanitize_filename

logger = logging.getLogger('types')

def _verify_dir(name: str, file_name: Path) -> bool:
    """
    verify a config directory exist. return false if verification fails
    """
    if file_name is not None and not file_name.is_dir():
        logger.error("Configured directory %s: %s is not a directory or not accessible", name, file_name)
        return False
    return True

def _verify_name_string(name: str, name_string: str) -> bool:
    """
    Verify the name format string.
    """
    info = LookedUpFileInfo()
    try:
        formatter = PartialFormatter()
        formatter.format(name_string, **info.asdict())
        return True
    except KeyError as key_error:
        logger.error("Configuration %s is not a valid file name format, please check %s", name, name_string)
        logger.error("Error message: %s", key_error)
        return False

class PartialFormatter(string.Formatter):
    """
    Used for formating NamerConfig.inplace_name and NamerConfig.
    """

    supported_keys = ['date','description','name','site','full_site','performers','all_performers','act','ext','trans']

    def __init__(self, missing='~~', bad_fmt='!!'):
        self.missing, self.bad_fmt=missing, bad_fmt

    def get_field(self, field_name, args, kwargs):
        # Handle a key not found
        try:
            val=super().get_field(field_name, args, kwargs)
        except (KeyError, AttributeError) as err:
            val=None,field_name
            if not field_name in self.supported_keys:
                raise KeyError(f"Key {field_name} not in support keys: {self.supported_keys}") from err
        return val

    def format_field(self, value, format_spec: str):
        if value is None:
            return self.missing
        try:
            if re.match(r'.\d+s', format_spec):
                value = value + format_spec[0] * int(format_spec[1:-1])
                format_spec = ''
            if re.match(r'.\d+p', format_spec):
                value = format_spec[0] * int(format_spec[1:-1]) + value
                format_spec = ''
            if re.match(r'.\d+i', format_spec):
                value = format_spec[0] * int(format_spec[1:-1]) + value + format_spec[0] * int(format_spec[1:-1])
                format_spec = ''
            return super().format_field(value, format_spec)
        except ValueError:
            if self.bad_fmt is not None:
                return self.bad_fmt
            raise

@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class NamerConfig():
    """
    Configuration for namer and namer_watchdog
    """

    # pylint: disable=too-many-instance-attributes

    porndb_token: str = None
    """
    token to access porndb.
    sign up here: https://metadataapi.net/
    """


    name_parser: str = '{_site}{_sep}{_date}{_sep}{_ts}{_name}{_dot}{_ext}'
    # pylint: disable=anomalous-backslash-in-string disable=line-too-long
    """
    This config may be a regex you provide, or a set of token used to build a regex.

    Supported token:

    ```
        _sep      r'[\.\- ]+'
        _site     r'(?P<site>[a-zA-Z0-9\.\-\ ]+[a-zA-Z0-9])'
        _date     r'(?P<year>[0-9]{2}(?:[0-9]{2})?)[\.\- ]+(?P<month>[0-9]{2})[\.\- ]+(?P<day>[0-9]{2})'
        _ts       r'((?P<trans>[T|t][S|s])'+_sep+'){0,1}'
        _name     r'(?P<name>.*)'
        _dot      r'\.'
        _ext      r'(?P<ext>[a-zA-Z0-9]{3,4})$'
    ```

    Full default regex:

    ```
    r'(?P<site>[a-zA-Z0-9\.\-\ ]+[a-zA-Z0-9])[\.\- ]+(?P<year>[0-9]{2}(?:[0-9]{2})?)[\.\- ]+(?P<month>[0-9]{2})[\.\- ]+(?P<day>[0-9]{2})((?P<trans>[T|t][S|s])[\.\- ]+){0,1}(?P<name>.*)\.(?P<ext>[a-zA-Z0-9]{3,4})$'
    ```

    The parts extracted from the file will be used in matching.
    _ts  is optional, and is mostly useful in stripping the marker from the _name, aiding in matching.
    _date if present is used to ensure the match is within 24 hours of your date.
    If the year/month/day capture groups are not present (due to you not using the supplied _date regex)
    dates will not be used in matching, possibly allowing false positives.
    """
    # pylint: enable=anomalous-backslash-in-string enable=line-too-long

    inplace_name: str = '{site} - {date} - {name}.{ext}'
    """
    How to write output file name.  When namer.py is run this is applied in place
    (next to the file to be processed).
    When run from namer_watchdog.py this will be written in to the successdir.

    Supports stand python 3 formating, as well as:
    * a prefix:
    {site: 1p} - in this case put one space in front of the 'site', so ' Vixen'
    * a suffix:
    {date:_1s} - in this case put one underscore after the 'date', so '2020-01-01_'
    * an infix:
    {date:_2i} - in this case put two underscore before and after the 'date', so '__2020-01-01__'

    Examples:
    * {site} - {date} - {scene}.{ext}
    * {site}/{date}.{scene}.{ext}

    Missing values will be ignored.

    Allowed replacements:

    * 'date' - in the format of YYYY-MM-DD.
    * 'description' - too long, don't use in a name.
    * 'name' - the scene name
    * 'site' - the site name, BrazzersExxtra, AllHerLuv, Deeper, etc with spaces removed.
    * 'full_site' - the site name from porndb, unmodified, i.e: Brazzers Exxtra, All Her Luv, etc.
    * 'performers' - space seperated list of female performers
    * 'all_performers' - space seperated list of all performers
    * 'act' - an act, parsed from original file name, don't use.
    * 'ext' - original file's extension, you should keep this.
    * 'trans' - 'TS', or 'ts' if detected in original file name.

    """

    min_file_size = 300
    """
    minimum file size to process in MB, ignored if a file is to be processed
    """

    prefer_dir_name_if_available = True
    """
    If a directory name is to be prefered over a file name.
    """

    set_uid = None
    """
    UID Settings for new/moved files/dirs.
    """

    set_gid = None
    """
    GID Settings for new/moved files/dirs.
    """

    write_namer_log = False
    """
    Should a log of comparisons be written next to processed video files.
    """

    set_dir_permissions = 775
    """
    Permissions Settings for new/moved dirs.
    """

    set_file_permissions = 664
    """
    Permissions Settings for new/moved file.
    """

    enabled_tagging: bool = True
    """
    Currently metadata pulled from the porndb can be added to mp4 files.
    This metadata will be read in fully by Plex, and Apple TV app, partially by Jellyfin (no artist support).
    Metadata includes, Title, Release Date, Scene Name, Artist, Source URL, XXX Movie rating.
    If a file is an mkv adding metadata at this time isn't supported.

    Should metadata fetched from the porn db be written in to the metadata of the mp4.
    No flags in this section will be used if this is not set to true.
    """

    enabled_poster: bool = True
    """
    Should the poster fetched from the porn db be written in to the metadata of the mp4.
    This poster will be displayed in Plex, Jellyfin and Apple TV app.
    Only applicable if enabled_tagging is True
    """

    enable_metadataapi_genres: bool = False
    """
    Should genres pulled from the porndb be added to the file?   These genres are noisey and
    not recommend for use.  If this is false a single default genere will be used.
    """

    default_genre: str = "Adult"
    """
    If genre's are not copied this is the default genere added to files.
    Default value is adult.
    """

    language: str = None
    """
    if language is set it will be used to select the default audio stream in an mp4 that has too many default stream
    to play correct in quicktime/apple tv.   If the language isn't found or is already the only default no action is
    taken, no streams (audio/video) are re-encoded.  Available here: https://iso639-3.sil.org/code_tables/639/data/
    """

    new_relative_path_name: str = '{site} - {date} - {name}/{site} - {date} - {name}.{ext}'
    """
    like inplace_name above used for local call for renaming, this instead is used to move a file/dir to a location relative
    to the dest_dir below on successful matching/tagging.
    """

    del_other_files = False
    """
    when processing a directory in the dest dir, should extra files besides the largest video be removed?
    or attempt to move them to a subdirectory specified in new_relative_path_name, if one exist.
    """

    watch_dir: Path = None
    """
    If running in watchdog mode, director where new downloads go.
    """

    work_dir: Path = None
    """
    If running in watchdog mode, temporary directory where work is done.
    a log file shows attempted matchs and match closeness.
    """

    failed_dir: Path = None
    """
    If running in watchdog mode, Should processing fail the file or directory is moved here.
    Files can be manually moved to watchdir to force reprocessing.
    """

    dest_dir: Path = None
    """
    If running in watchdog mode, dir where finalized files get written.
    """

    retry_time: str = None
    """
    Time to retry failed items every day.
    """

    extra_sleep_time: int = 30
    """
    Extra time to sleep in seconds to allow all information to be copied in dir
    """

    def __init__(self):
        if sys.platform != "win32":
            self.set_uid = os.getuid()
            self.set_gid = os.getgid()


    def __str__(self):
        token = "None In Set, Go to https://metadatapi.net/ to get one!"
        if self.porndb_token is not None:
            token = re.sub(r'.', '*', self.porndb_token)
        output = 'Namer Config:\n'
        output += f'porndb_token: {token}\n'
        output += f"  inplace_name: {self.inplace_name}\n"
        output += f"  prefer_dir_name_if_available: {self.prefer_dir_name_if_available}\n"
        output += f"  set_uid: {self.set_uid}\n"
        output += f"  set_gid: {self.set_gid}\n"
        output += f"  write_namer_log: {self.write_namer_log}\n"
        output += f"  set_dir_permissions: {self.set_dir_permissions}\n"
        output += f"  set_file_permissions: {self.set_file_permissions}\n"
        output += "Tagging Config:\n"
        output += f"  enabled_tagging: {self.enabled_tagging}\n"
        output += f"  enabled_poster: {self.enabled_poster}\n"
        output += f"  enable_metadataapi_genres: {self.enable_metadataapi_genres}\n"
        output += f"  default_genre: {self.default_genre}\n"
        output += f"  language: {self.language}\n"
        output += "Watchdog Config:\n"
        output += f"  min_file_size: {self.min_file_size}mb\n"
        output += f"  del_other_files: {self.del_other_files}\n"
        output += f"  new_relative_path_name: {self.new_relative_path_name}\n"
        output += f"  watch_dir: {self.watch_dir}\n"
        output += f"  work_dir: {self.work_dir}\n"
        output += f"  failed_dir: {self.failed_dir}\n"
        output += f"  dest_dir: {self.dest_dir}\n"
        output += f"  retry_time: {self.retry_time}\n"
        output += f"  extra_sleep_time: {self.extra_sleep_time}\n"
        return output


    def verify_naming_config(self) -> bool:
        """
        Verifies the contents of your config file. Returns False if configuration failed.
        """
        success = True
        if self.enable_metadataapi_genres is not True and self.default_genre is None:
            logger.error("Sinse enable_metadataapi_genres is not True, you must specify a default_genre")
            success = False
        success = _verify_name_string("inplace_name", self.inplace_name) and success
        return success


    def verify_watchdog_config(self) -> bool:
        """
        Verifies the contents of your config file. Returns False if configuration failed.
        """
        success = True
        if self.enable_metadataapi_genres is not True and self.default_genre is None:
            logger.error("Sinse enable_metadataapi_genres is not True, you must specify a default_genre")
            success = False
        success = _verify_dir("watch_dir", self.watch_dir) and success
        success = _verify_dir("work_dir", self.work_dir) and success
        success = _verify_dir("failed_dir", self.failed_dir) and success
        success = _verify_dir("dest_dir", self.dest_dir) and success
        success = _verify_name_string("new_relative_path_name", self.new_relative_path_name) and success
        return success


def from_config(config : ConfigParser) -> NamerConfig:
    """
    Given a config parser pointed at a namer.cfg file, return a NamerConfig with the file's parameters.
    """
    namer_config = NamerConfig()
    namer_config.porndb_token = config.get('namer','porndb_token', fallback=None)
    namer_config.inplace_name = config.get('namer','inplace_name',fallback='{site} - {date} - {name}.{ext}')
    namer_config.name_parser = config.get('namer','name_parser',fallback='{_site}{_sep}{_date}{_sep}{_ts}{_name}{_dot}{_ext}')
    namer_config.prefer_dir_name_if_available = config.getboolean('namer','prefer_dir_name_if_available',fallback=False)
    namer_config.min_file_size = config.getint('namer','min_file_size',fallback=100)
    namer_config.set_uid = config.getint('namer','set_uid',fallback=None)
    namer_config.set_gid = config.getint('namer','set_gid',fallback=None)
    namer_config.write_namer_log = config.getboolean('namer','write_namer_log',fallback=False)
    namer_config.set_dir_permissions = config.get('namer','set_dir_permissions',fallback=775)
    namer_config.set_file_permissions = config.get('namer','set_file_permissions',fallback=664)
    namer_config.enabled_tagging = config.getboolean('metadata','enabled_tagging',fallback=True)
    namer_config.enabled_poster = config.getboolean('metadata','enabled_poster',fallback=True)
    namer_config.enable_metadataapi_genres = config.getboolean('metadata','enable_metadataapi_genres',fallback=False)
    namer_config.default_genre = config.get('metadata','default_genre',fallback='Adult')
    namer_config.language = config.get('metadata','language',fallback=None)
    namer_config.new_relative_path_name = config.get('watchdog','new_relative_path_name',
        fallback='{site} - {date} - {name}/{site} - {date} - {name}.{ext}')
    namer_config.del_other_files = config.getboolean('watchdog','del_other_files',fallback=False)
    namer_config.extra_sleep_time = config.getint('watchdog','extra_sleep_time',fallback=30)
    watch_dir = config.get('watchdog','watch_dir',fallback=None)
    if watch_dir is not None:
        namer_config.watch_dir = Path(watch_dir)
    work_dir = config.get('watchdog','work_dir',fallback=None)
    if work_dir is not None:
        namer_config.work_dir = Path(work_dir)
    failed_dir = config.get('watchdog','failed_dir',fallback=None)
    if failed_dir is not None:
        namer_config.failed_dir = Path(failed_dir)
    dest_dir = config.get('watchdog','dest_dir',fallback=None)
    if dest_dir is not None:
        namer_config.dest_dir = Path(dest_dir)

    namer_config.retry_time = config.get('watchdog','retry_time',fallback=None)
    if namer_config.retry_time is None:
        namer_config.retry_time = f"03:{random.randint(0, 59):0>2}"
    return namer_config

def default_config() -> NamerConfig:
    """
    Attempts reading various locations to fine a namer.cfg file.
    """
    config = configparser.ConfigParser()
    default_locations = [Path.home() / ".namer.cfg", Path('./namer.cfg')]
    if os.environ.get('NAMER_CONFIG') is not None:
        default_locations.insert(0, Path(os.environ.get('NAMER_CONFIG')))
    config.read(default_locations)
    return from_config(config)

@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class FileNameParts:
    """
    Represents info parsed from a file name, usually of an nzb, named something like:
    'EvilAngel.22.01.03.Carmela.Clutch.Fabulous.Anal.3-Way.XXX.2160p.MP4-GAYME-xpost'
    or
    'DorcelClub.20.12..Aya.Benetti.Megane.Lopez.And.Bella.Tina.2160p.MP4-GAYME-xpost'
    """

    # pylint: disable=too-many-instance-attributes

    site: str = None
    """
    Site the file originated from, "DorcelClub", "EvilAngel", etc.
    """
    date: str = None
    """
    formated: YYYY-mm-dd
    """
    trans: bool = False
    """
    If the name originally started with an "TS" or "ts"
    it will be stripped out and placed in a seperate location, aids in matching, useable to genre mark content.
    """
    name: str = None
    act: str = None
    """
    If the name originally ended with an "act ###" or "part ###"
    it will be stripped out and placed in a seperate location, aids in matching.
    """
    extension: str = None
    """
    The file's extension .mp4 or .mkv
    """
    resolution: str = None
    """
    Resolution, if the file name makes a claim about resolution. (480p, 720p, 1080p, 4k)
    """
    source_file_name: str = None
    """
    What was originally parsed.
    """

    def __str__(self) -> str:
        return f"""site: {self.site}
        date: {self.date}
        trans: {self.trans}
        name: {self.name}
        act: {self.act}
        extension: {self.extension}
        original full name: {self.source_file_name}
        """

@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class Performer:
    """
    Minimal info about a perform, name, and role.
    """
    name: str
    role: str
    """
    if available the performers gender, stored as a role.  example: "Female", "Male"
    Useful as many nzbs often don't include the scene name, but instead female performers names,
    or sometimes both.
    Other performers are also used in name matching, if females are attempted first.
    """

    def __init__(self, name=None, role=None):
        self.name = name
        self.role = role

    def __str__(self):
        name = "Unknown" if self.name is None else self.name
        if self.role is not None:
            return name + " (" + self.role + ")"
        return name

    def __repr__(self):
        return f'Performer[name={self.name}, role={self.role}]'


@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class LookedUpFileInfo():
    """
    Information from a call to the porndb about a specific scene.
    """

    # pylint: disable=too-many-instance-attributes

    uuid: str = ""
    """
    porndb scene id, allowing lookup of more metadata, (tags)
    """

    site: str = ""
    """
    Site where this video originated, DorcelClub/Deeper/etc.....
    """
    date: str = ""
    """
    date of initial release, formated YYYY-mm-dd
    """
    name: str = ""
    """
    Name of the scene in this video
    """
    description: str = ""
    """
    Description of the action in this video
    """
    source_url: str = ""
    """
    Original source location of this video
    """
    poster_url: str = ""
    """
    Url to download a poster for this video
    """
    performers: List[Performer]
    """
    List of performers, containing names, and "roles" aka genders, for each performer.
    """
    genres: List[str]
    """
    List of genres, per porndb.  Tends to be noisey.
    """
    origninal_response: str = ""
    """
    json reponse parsed in to this object.
    """
    original_query: str = ""
    """
    url query used to get the above json response
    """
    original_parsed_filename: FileNameParts
    """
    The FileNameParts used to build the orignal_query
    """
    look_up_site_id: str = ""
    """
    ID Used by the queried site to identify the video
    """

    tags: List[str]
    """
    Tags associated with the video.   Noisy and long list.
    """

    def __init__(self):
        self.performers = []
        self.tags = []
        self.original_parsed_filename = FileNameParts()

    def asdict(self):
        """
        Converts the info in to a dict that can be used
        by PartialFormatter to return a new path for a file.
        """
        if self.original_parsed_filename is None:
            self.original_parsed_filename = FileNameParts()
        return {'uuid': self.uuid,
                'date': self.date,
                'description': self.description,
                'name': self.name,
                'site': self.site.replace(' ',''),
                'full_site': self.site,
                'performers': " ".join(map(lambda p: p.name, filter( lambda p: p.role == 'Female' , self.performers))),
                'all_performers': " ".join(map( lambda p: p.name , self.performers)),
                'act': self.original_parsed_filename.act,
                'ext': self.original_parsed_filename.extension,
                'trans': self.original_parsed_filename.trans}

    def new_file_name(self, template: str, infix: str = "(0)") -> str:
        """
        Constructs a new file name based on a template (describe in NamerConfig)
        """
        dictionary = self.asdict()
        clean_dic = { k: str(sanitize_filename(str(v), platform=Platform.UNIVERSAL))  for k, v in dictionary.items() }
        fmt = PartialFormatter(missing="", bad_fmt="---")
        name = fmt.format(template, **clean_dic)
        if infix != str("(0)"):
            # will apply the infix before the file extension if just a file name, if a path, with apply
            #the infix after the fist part (first directory name) of the (sub)path
            path = PurePath(name)
            if len(path.parts) > 1:
                name = str(path.parent / ( path.stem + infix + path.suffix ))
            else:
                name = path.stem + infix + path.suffix
        return name

@dataclass(init=True, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class ComparisonResult:
    """
    Represents the comparison from a FileNameParts and a LookedUpFileInfo, it will be
    consider a match if the creation dates match, the studio matches, and the original
    scene/perform part of a the file name can match any combination of the metadata about
    actor names and/or scene name.   RapidFuzz is used to make the comparison.
    """

    name: str
    name_match: float
    """
    How closely did the name found in FileNameParts match (via RapidFuzz string comparison)
    The performers and scene name found in LookedUpFileInfo.  Various combinations of performers
    and scene namer are used for attempted matching.
    """

    sitematch: bool
    """
    Did the studios match between filenameparts and lookedup
    """


    datematch: bool
    """
    Did the dates match between filenameparts and lookedup
    """

    name_parts: FileNameParts
    """
    Parts of the file name that were parsed and used as search parameters.
    """

    looked_up: LookedUpFileInfo
    """
    Info pulled from the porndb.  When doing searchs it will not include tags, only included when
    performing a lookup by id (which is done only after a match is made.)
    """

    def is_match(self) -> bool:
        """
        Returns true if site and creation data match exactly, and if the name fuzzes against
        the metadate to 90% or more (via RapidFuzz, and various concatinations of metadata about
        actors and scene name).
        """
        return self.sitematch and self.datematch and self.name_match >= 89.9



@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class ProcessingResults:
    """
    Returned from the namer.py's process() function.   It contains information about if a match
    was found, and of so, where files were placed.  It also tracks if a directory was inputed
    to namer (rather than the exact movie file.)  That knowledge can be used to move directories
    and preserve relative files, or to delete left over artifacts.
    """

    search_results: List[ComparisonResult] = None
    """
    True if a match was found in the porndb.
    """

    new_metadata: LookedUpFileInfo = None
    """
    New metadata found for the file being processed.
    Sourced including queries against the porndb, which would be stored in search_results,
    or reading a .nfo xml file next to the video, with the file name identical exept for
    the extension, which would be .nfo instead of .mkv, or .mp4.
    """

    dirfile: Path = None
    """
    Set if the input file for naming was a directory.   This has advantages, as clean up of other files is now possible,
    or all files can be moved to a destination specified in the field final_name_relative.
    """

    video_file: Path = None
    """
    The location of the found video file.
    """

    parsed_file: FileNameParts = None
    """
    The parsed file name.
    """

    final_name_relative: str = None
    """
    This is the full NamerConfig.new_relative_path_name string with all substitutions made.
    """

    namer_log_file: Path = None
    """
    If a log file was written, where it was stored.
    """



"""
ffmpeg is access through this file, it is used to find the video streams resolution,
and update audio streams "Default" setting.   Apple video players require there be
only one default audio stream, and this script lets you set it with the correct language
code their are more than one audio streams and if they are correctly labeled.
See:  https://iso639-3.sil.org/code_tables/639/data/
"""

import json
from types import SimpleNamespace
import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
import logging

logger = logging.getLogger('ffmpeg')


def get_resolution(file: str) -> int:
    """
    Gets the vertical resolution of an mp4 file.  For example, 720, 1080, 2160...
    """
    logger.info("resolution stream of file %s", file)

    with subprocess.Popen(['ffprobe',
                           '-v',
                           'error',
                           '-select_streams',
                           'v:0',
                           '-show_entries',
                           'stream=height',
                           '-of',
                           'csv=p=0',
                           file],
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          universal_newlines=True) as process:
        success = process.wait() == 0
        output = process.stdout.read()
        if not success:
            logger.warning("Error gettng resolution of file %s", file)
            logger.warning(process.stderr.read())
        process.stdout.close()
        process.stderr.close()
        logger.info("output %s", output)
        return int(output)
    return None


def get_audio_stream_for_lang(mp4_file: str, language: str) -> int:
    """
    given an mp4 input file and a desired language will return the stream position of that language in the mp4.
    if the language is None, or the stream is not found, or the desired stream is the only default stream, None is returned.
    See: https://iso639-3.sil.org/code_tables/639/data/
    """

    with subprocess.Popen(['ffprobe',
                           '-show_streams',
                           '-select_streams',
                           'a',
                           '-of',
                           'json',
                           '-i',
                           mp4_file],
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          universal_newlines=True) as process:
        success = process.wait() == 0
        audio_streams_str = process.stdout.read()
        if not success:
            logger.warning("Error gettng audio streams of file %s", input)
            logger.warning(process.stderr.read())
        process.stdout.close()
        process.stderr.close()

        logger.info("Target for audio: %s", input)

        audio_streams = json.loads(
            audio_streams_str, object_hook=lambda d: SimpleNamespace(**d))
        lang_stream = None
        needs_updated = False
        if language:
            test_lang = language.lower()[0:3]
            if audio_streams is not None and hasattr(audio_streams, 'streams'):
                for audio_stream in audio_streams.streams:
                    default = audio_stream.disposition.default == 1
                    lang = audio_stream.tags.language
                    if lang == test_lang:
                        lang_stream = audio_stream.index - 1
                        if default is False:
                            needs_updated = True
                    elif default is True:
                        needs_updated = True
                if needs_updated and lang_stream:
                    return lang_stream
    return None


def update_audio_stream_if_needed(mp4_file: Path, language: str) -> bool:
    """
    Returns true if the file had to be edited to have a default audio stream equal to the desired language,
    mostly a concern for apple players (Quicktime/Apple TV/etc.)
    Copies, and potentially updates the default audio stream of a video file.
    """

    with TemporaryDirectory() as tempdir:
        workfile = Path(tempdir) / mp4_file.name
        stream = get_audio_stream_for_lang(mp4_file, language)
        if stream is not None:
            logger.info(
                "Attempt to alter default audio stream of %s", mp4_file)
            with subprocess.Popen(['ffmpeg',
                                   '-i',
                                   mp4_file,  # input file
                                   '-map',
                                   '0',  # copy all stream
                                   '-disposition:a',
                                   'none',  # mark all audio streams as not default
                                   # mark this audio stream as default
                                   f'-disposition:a:{stream}',
                                   'default',
                                   '-c',
                                   'copy',  # don't reencode anything.
                                   workfile],
                                  stdout=subprocess.DEVNULL,
                                  stderr=subprocess.PIPE,
                                  universal_newlines=True) as process:
                stderr = process.stderr.read()
                success = process.wait() == 0
                process.stderr.close()
                if not success:
                    logger.info(
                        "Could not update audio stream for %s", mp4_file)
                    logger.info(stderr)
                else:
                    logger.warning("Return code: %s", process.returncode)
                    mp4_file.unlink()
                    workfile.rename(mp4_file)
                return success
    return True
